import base64
import hashlib
from typing import Final

from cryptography.fernet import Fernet, InvalidToken

ROOT_ENCRYPTION_KEY_ENV_VAR: Final[str] = "ROOT_ENCRYPTION_KEY"


class RootSecretError(Exception):
    """Base exception for package errors."""


class MissingRootEncryptionKeyError(RootSecretError):
    """Raised when a required root encryption key is not available."""


class InvalidRootEncryptionKeyError(RootSecretError):
    """Raised when a root encryption key cannot initialize Fernet."""


class SecretDecryptionError(RootSecretError):
    """Raised when a secret cannot be decrypted."""


def generate_root_encryption_key() -> str:
    return Fernet.generate_key().decode("ascii")


def normalize_root_encryption_key(raw_key: str) -> str:
    key = raw_key.strip()
    if not key:
        raise InvalidRootEncryptionKeyError("Root encryption key cannot be empty.")

    try:
        Fernet(key.encode("ascii"))
        return key
    except (UnicodeEncodeError, ValueError):
        pass

    derived_key = base64.urlsafe_b64encode(hashlib.sha256(key.encode("utf-8")).digest())
    try:
        Fernet(derived_key)
    except ValueError as exc:
        raise InvalidRootEncryptionKeyError("Root encryption key is invalid.") from exc
    return derived_key.decode("ascii")


def build_fernet(raw_key: str) -> Fernet:
    normalized_key = normalize_root_encryption_key(raw_key)
    return Fernet(normalized_key.encode("ascii"))


def encrypt_value(value: str, raw_key: str) -> str:
    return build_fernet(raw_key).encrypt(value.encode("utf-8")).decode("ascii")


def decrypt_value(encrypted_value: str, raw_key: str) -> str:
    try:
        decrypted = build_fernet(raw_key).decrypt(encrypted_value.encode("ascii"))
    except InvalidToken as exc:
        raise SecretDecryptionError("Encrypted value could not be decrypted with the root encryption key.") from exc
    except UnicodeEncodeError as exc:
        raise SecretDecryptionError("Encrypted value must be ASCII-safe text.") from exc
    return decrypted.decode("utf-8")



from django.core.management.base import BaseCommand, CommandError

from django_root_secret.crypto import generate_root_encryption_key
from django_root_secret.env import ensure_path_is_gitignored, env_file_path, write_root_key_file


class Command(BaseCommand):
    help = "Create an <env>.env file containing a single ROOT_ENCRYPTION_KEY."

    def add_arguments(self, parser):
        parser.add_argument("--env", required=True, help="Environment name used for the <env>.env file.")

    def handle(self, *args, **options):
        try:
            path = env_file_path(options["env"])
        except ValueError as exc:
            raise CommandError(str(exc)) from exc

        root_key = generate_root_encryption_key()
        try:
            write_root_key_file(path, root_key)
        except FileExistsError as exc:
            raise CommandError(f"Environment file already exists: {path}") from exc

        gitignore_updated = ensure_path_is_gitignored(path)

        self.stdout.write(f"Created root encryption key file: {path}")
        if gitignore_updated:
            self.stdout.write(f"Added {path.name} to {path.parent / '.gitignore'}")
        else:
            self.stdout.write(f"{path.name} is already present in {path.parent / '.gitignore'}")



from django.core.management.base import BaseCommand, CommandError

from django_root_secret.crypto import (
    InvalidRootEncryptionKeyError,
    MissingRootEncryptionKeyError,
    encrypt_value,
)
from django_root_secret.env import env_file_path, load_root_key_from_env_file


class Command(BaseCommand):
    help = "Encrypt a plaintext value using ROOT_ENCRYPTION_KEY from <env>.env."

    def add_arguments(self, parser):
        parser.add_argument("--value", required=True, help="Plaintext value to encrypt.")
        parser.add_argument("--env", required=True, help="Environment name used for the <env>.env file.")

    def handle(self, *args, **options):
        try:
            path = env_file_path(options["env"])
            root_key = load_root_key_from_env_file(path)
        except (ValueError, MissingRootEncryptionKeyError, InvalidRootEncryptionKeyError) as exc:
            raise CommandError(str(exc)) from exc

        encrypted_value = encrypt_value(options["value"], root_key)
        self.stdout.write(encrypted_value)



from django.apps import AppConfig


class DjangoRootSecretConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_root_secret"
    verbose_name = "Django Root Secret"



import os
from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import SimpleTestCase

from django_root_secret.crypto import MissingRootEncryptionKeyError, SecretDecryptionError, encrypt_value
from django_root_secret.secrets import get_secret


class GenerateRootEncryptionKeyCommandTests(SimpleTestCase):
    def test_creates_env_file_and_prints_key_and_location(self):
        with TemporaryDirectory() as temp_dir:
            output = StringIO()
            with mock.patch("pathlib.Path.cwd", return_value=Path(temp_dir)):
                call_command("generate_root_encryption_key", env="development", stdout=output)

            env_file = Path(temp_dir) / "development.env"
            self.assertTrue(env_file.exists())
            content = env_file.read_text(encoding="utf-8")
            self.assertIn("ROOT_ENCRYPTION_KEY=", content)
            rendered_output = output.getvalue()
            self.assertTrue(rendered_output.strip())
            self.assertIn(str(env_file), rendered_output)
            self.assertIn("Added development.env to", rendered_output)

    def test_adds_env_file_to_gitignore_when_missing(self):
        with TemporaryDirectory() as temp_dir:
            output = StringIO()
            gitignore_path = Path(temp_dir) / ".gitignore"
            gitignore_path.write_text(".venv/\n", encoding="utf-8")

            with mock.patch("pathlib.Path.cwd", return_value=Path(temp_dir)):
                call_command("generate_root_encryption_key", env="development", stdout=output)

            self.assertIn("development.env\n", gitignore_path.read_text(encoding="utf-8"))
            self.assertIn("Added development.env to", output.getvalue())

    def test_reports_when_env_file_is_already_gitignored(self):
        with TemporaryDirectory() as temp_dir:
            output = StringIO()
            gitignore_path = Path(temp_dir) / ".gitignore"
            gitignore_path.write_text("development.env\n", encoding="utf-8")

            with mock.patch("pathlib.Path.cwd", return_value=Path(temp_dir)):
                call_command("generate_root_encryption_key", env="development", stdout=output)

            self.assertIn("already present", output.getvalue())

    def test_creates_env_file_with_private_permissions(self):
        with TemporaryDirectory() as temp_dir:
            with mock.patch("pathlib.Path.cwd", return_value=Path(temp_dir)):
                call_command("generate_root_encryption_key", env="development")

            env_file = Path(temp_dir) / "development.env"
            self.assertEqual(env_file.stat().st_mode & 0o777, 0o600)

    def test_raises_when_env_file_exists(self):
        with TemporaryDirectory() as temp_dir:
            env_file = Path(temp_dir) / "development.env"
            env_file.write_text("ROOT_ENCRYPTION_KEY=existing\n", encoding="utf-8")

            with mock.patch("pathlib.Path.cwd", return_value=Path(temp_dir)):
                with self.assertRaises(CommandError):
                    call_command("generate_root_encryption_key", env="development")


class EncryptSecretCommandTests(SimpleTestCase):
    def test_encrypts_value_with_env_file_key(self):
        with TemporaryDirectory() as temp_dir:
            call_command_output = StringIO()
            with mock.patch("pathlib.Path.cwd", return_value=Path(temp_dir)):
                call_command("generate_root_encryption_key", env="production")
                call_command(
                    "encrypt_secret",
                    env="production",
                    value="super-secret",
                    stdout=call_command_output,
                )

            encrypted_value = call_command_output.getvalue().strip()
            env_file = Path(temp_dir) / "production.env"
            root_key = [
                line.split("=", 1)[1]
                for line in env_file.read_text(encoding="utf-8").splitlines()
                if line.startswith("ROOT_ENCRYPTION_KEY=")
            ][0]
            self.assertEqual(get_secret_with_key(encrypted_value, root_key), "super-secret")

    def test_raises_when_env_file_is_missing(self):
        with TemporaryDirectory() as temp_dir:
            with mock.patch("pathlib.Path.cwd", return_value=Path(temp_dir)):
                with self.assertRaises(CommandError):
                    call_command("encrypt_secret", env="missing", value="value")


class GetSecretTests(SimpleTestCase):
    def test_decrypts_using_root_encryption_key_env_var(self):
        encrypted = encrypt_value("plain-value", "not-a-fernet-key")
        with mock.patch.dict(os.environ, {"ROOT_ENCRYPTION_KEY": "not-a-fernet-key"}, clear=False):
            self.assertEqual(get_secret(encrypted), "plain-value")

    def test_raises_when_root_key_env_var_is_missing(self):
        with mock.patch.dict(os.environ, {}, clear=True):
            with self.assertRaises(MissingRootEncryptionKeyError):
                get_secret("token")

    def test_raises_when_secret_cannot_be_decrypted(self):
        encrypted = encrypt_value("plain-value", "first-key")
        with mock.patch.dict(os.environ, {"ROOT_ENCRYPTION_KEY": "second-key"}, clear=False):
            with self.assertRaises(SecretDecryptionError):
                get_secret(encrypted)


def get_secret_with_key(encrypted_value: str, root_key: str) -> str:
    with mock.patch.dict(os.environ, {"ROOT_ENCRYPTION_KEY": root_key}, clear=False):
        return get_secret(encrypted_value)



import os
from pathlib import Path

from .crypto import MissingRootEncryptionKeyError

ROOT_ENCRYPTION_KEY_FILE_VAR = "ROOT_ENCRYPTION_KEY"
FILE_COMMENT = (
    "# This file must only contain ROOT_ENCRYPTION_KEY.\n"
    "# Encrypt every other secret with this key and keep the file private.\n"
)


def validate_env_name(env_name: str) -> str:
    normalized = env_name.strip()
    if not normalized:
        raise ValueError("Environment name cannot be empty.")

    allowed = set("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-")
    if any(char not in allowed for char in normalized):
        raise ValueError(
            "Environment name may only contain letters, numbers, underscores, and hyphens."
        )
    return normalized


def env_file_path(env_name: str, base_dir: Path | None = None) -> Path:
    normalized = validate_env_name(env_name)
    directory = Path.cwd() if base_dir is None else Path(base_dir)
    return directory / f"{normalized}.env"


def ensure_path_is_gitignored(path: Path, gitignore_path: Path | None = None) -> bool:
    target_gitignore = path.parent / ".gitignore" if gitignore_path is None else Path(gitignore_path)
    entry = path.name

    if target_gitignore.exists():
        lines = target_gitignore.read_text(encoding="utf-8").splitlines()
    else:
        lines = []

    normalized_entries = {line.strip() for line in lines}
    if entry in normalized_entries:
        return False

    with target_gitignore.open("a", encoding="utf-8") as file_obj:
        if lines and lines[-1].strip():
            file_obj.write("\n")
        file_obj.write(f"{entry}\n")
    return True


def write_root_key_file(path: Path, root_key: str) -> None:
    content = f"{FILE_COMMENT}{ROOT_ENCRYPTION_KEY_FILE_VAR}={root_key}\n"
    flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
    fd = os.open(path, flags, 0o600)
    try:
        with os.fdopen(fd, "w", encoding="utf-8") as file_obj:
            file_obj.write(content)
    except Exception:
        path.unlink(missing_ok=True)
        raise


def load_root_key_from_env_file(path: Path) -> str:
    if not path.exists():
        raise MissingRootEncryptionKeyError(f"Environment file does not exist: {path}")

    for raw_line in path.read_text(encoding="utf-8").splitlines():
        line = raw_line.strip()
        if not line or line.startswith("#"):
            continue
        if "=" not in line:
            continue
        name, value = line.split("=", 1)
        if name.strip() != ROOT_ENCRYPTION_KEY_FILE_VAR:
            continue
        cleaned = value.strip().strip("\"'")
        if not cleaned:
            break
        return cleaned

    raise MissingRootEncryptionKeyError(
        f"{ROOT_ENCRYPTION_KEY_FILE_VAR} is missing from environment file: {path}"
    )
